
/* Netfilter redirector v1.0 by Marius Petrescu, YO2LOJ <marius@yo2loj.ro>
 * Redirect all incoming IPIP packets to protocol IPPROTO_DIVERT */

#include <linux/version.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/skbuff.h>
#include <linux/ip.h>
#include <linux/netfilter.h>
#include <linux/if_ether.h>
#include <linux/netfilter_ipv4.h>

/* we will use proto 94 (IPIP) to intercept proto 4 (IPENCAP) */
#define IPPROTO_DIVERT       94


/* This is the structure we shall use to register our functions */
static struct nf_hook_ops nfho_in;
static struct nf_hook_ops nfho_out;

/* These are the hook filter functions - prototype has changed in kernel version 4 */

/* input hook function */
static unsigned int hook_func_in(void *priv, struct sk_buff *skb, const struct nf_hook_state *state)
{
    struct iphdr *iph = ip_hdr(skb);
    if (iph->protocol == IPPROTO_IPIP)
    {
        iph->protocol = IPPROTO_DIVERT;

        skb->ip_summed = CHECKSUM_NONE;
        skb->csum_valid = 0;
        iph->check = 0;
        iph->check = ip_fast_csum((u8 *)iph, iph->ihl);

        //printk("IPIP redirector: Incoming IPIP packet intercepted\n");
    }
    return NF_ACCEPT;
}

/* output hook function */
static unsigned int hook_func_out(void *priv, struct sk_buff *skb, const struct nf_hook_state *state)
{
    struct iphdr *iph = ip_hdr(skb);
    if (iph->protocol == IPPROTO_DIVERT)
    {
        iph->protocol = IPPROTO_IPIP;

        skb->ip_summed = CHECKSUM_NONE;
        skb->csum_valid = 0;
        iph->check = 0;
        iph->check = ip_fast_csum((u8 *)iph, iph->ihl);

        //printk("IPIP redirector: Outgoing DIVERT packet intercepted\n");
    }
    return NF_ACCEPT;
}

/* Initialisation routine */
static int __init init_nf(void)
{
    /* Fill in our hook structures */
    nfho_in.hook = hook_func_in;
    nfho_in.hooknum = NF_INET_LOCAL_IN;
    nfho_in.pf = PF_INET;
    nfho_in.priority = NF_IP_PRI_LAST;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,13,0)
    nf_register_net_hook(&init_net, &nfho_in);
#else
    nf_register_hook(&nfho_in);
#endif

    nfho_out.hook = hook_func_out;
    nfho_out.hooknum = NF_INET_LOCAL_OUT;
    nfho_out.pf = PF_INET;
    nfho_out.priority = NF_IP_PRI_FIRST;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,13,0)
    nf_register_net_hook(&init_net, &nfho_out);
#else
    nf_register_hook(&nfho_out);
#endif

    printk("IPIP Redirector - part of AMPRD by Marius Petrescu, YO2LOJ\n");
    printk("Netfilter redirector registered\n");

    return 0;
}

/* Cleanup routine */
static void __exit exit_nf(void)
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,13,0)
    nf_unregister_net_hook(&init_net, &nfho_in);
    nf_unregister_net_hook(&init_net, &nfho_out);
#else
    nf_unregister_hook(&nfho_in);
    nf_unregister_hook(&nfho_out);
#endif
    printk("IPIP Netfilter redirector unregistered\n");
}

module_init(init_nf);
module_exit(exit_nf);
MODULE_LICENSE("GPL");
MODULE_AUTHOR("Marius Petrescu YO2LOJ <marius@yo2loj.ro>");
